/*****************************************************************************\
 *
 * tanky.h
 *
 * Tank Hunters
 *
 * Authors: Daniel Vychopen (xvycho04 _AT stud.fit.vutbr.cz)
 *
 * Contributors:
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#if defined(_WIN32) || defined(__WIN32__)
#pragma comment(linker, "/entry:\"mainCRTStartup\"") // Pro linker: Budeme linkovat s main,
                                                 // nikoliv WinMain, jak je na Windows zvykem
#pragma comment(linker, "/subsystem:\"console\"") //zapnout konzoli

# include <Inventor/Win/SoWin.h>
# include <Inventor/Win/SoWinRenderArea.h>
#else
# include <Inventor/Qt/SoQt.h>
# include <Inventor/Qt/SoQtRenderArea.h>
#endif

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoFile.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoTransparencyType.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/sensors/SoOneShotSensor.h>
#include <Inventor/collision/SoIntersectionDetectionAction.h>
#include <Inventor/VRMLnodes/SoVRMLImageTexture.h>


const float maxv=50.0f; // maximalni rychlost tanku
const float minv=-15.0f; // minimalni rychlost tanku
const float zrychleni=15.0f; // zrychleni tanku
const float zpomaleni=50.0f; // zpomaleni tanku
const float maxvOt=1.5;  //maximalni rychlost otaceni tanku
const float minvOt=-1.5;  //minimalni rychlost otaceni tanku
const float zrychleniOt=6;//zrychleni otaceni tanku
const float zpomaleniOt=30;//zpomaleni otaceni tanku
const float maxSklonHlavne=0.7f; // maximalni sklon hlavne tanku
const float minSklonHlavne=0.0f; // minimalni sklon hlavne tanku
const float zrychleniSklonu=3;  //zrychleni sklonu hlavne tanku
const float zpomaleniSklonu=30; //zpomaleni sklonu hlavne tanku
const float maxvSklonu=0.5;  //maximalni rychlost sklanini hlavne tanku
const float minvSklonu=-0.5;  //minimalni rychlost sklanini hlavne tanku
const float zpomaleniZed=100;//zpomaleni o zed
const float gravitace=500.f;//gravitace


SoIntersectionDetectionAction *ida;//testovani kolizi tanku s bludistem
SoIntersectionDetectionAction *ida2;//testovani kolizi strely s bludistem a tanky
SoIntersectionDetectionAction *ida3;//testovani kolizi mezi tanky

SoSeparator *rootkolize; // koren pro pocitani kolizi
SoSeparator *bludisteKolize;
bool kolize=false;
float tmpangle;


// koren sceny
SoSeparator *root;

// kamera
SoPerspectiveCamera *kamera;


class Tank
{
  float natoceni;
  float rychlost;    //rychlost tanku
  SoTranslation *trans;
  SoRotation *rotace;
  SoSeparator *root;
  SoFile *model;
public:
  int skore;
  float sklonHlavne;
  SbVec3f minulaPozice; // pokud tanky vjedou na stejne misto ve stejny okamzik, musi se vratit
  float minuleNatoceni;
  float getSklonHlavne() {return sklonHlavne;};
  SoSeparator *rootKolize;
  void init(SoSeparator *root, SoSeparator *rootKolize);
  SbVec3f smer;
  float KorekceUhlu;
  void setRychlost(float v) {rychlost=v;}
  float getRychlost() const {return rychlost;}
  float rychlostOt;  //rychlost otaceni tanku
  float rychlostSklonu;//rychlost sklonu hlavne tanku
    SbVec3f getPosition() const { return trans->translation.getValue();}
  void getPosition(float &x, float &y, float &z) const {trans->translation.getValue().getValue(x, y, z); }
  void setPosition(const SbVec3f pozice) {trans->translation.setValue(pozice);}
    void setPosition(const float x, const float y, const float z) {trans->translation.setValue(x, y, z);}
  float getRotation() const{return natoceni;}
  void setRotation(const float & uhel) {natoceni=uhel; rotace->rotation.setValue(SbVec3f(0, 1, 0), uhel);}
  SoFile *modelKolize;
} tank;


void Tank::init(SoSeparator *root, SoSeparator *rootKolize)
{
  skore=0;
  rychlost=0;
  rychlostOt=0;
  rychlostSklonu=0;
  sklonHlavne=0.128f;

  // tank
  this->root = new SoSeparator;
  root->addChild(this->root);

  // posunuti tanku
  this->trans = new SoTranslation;
  this->root->addChild(this->trans);

  // otoceni tanku
  this->rotace = new SoRotation;
  this->root->addChild(this->rotace);


  // nacteni modelu tanku
  this->model = new SoFile;
  this->model->name.setValue("models/tank.wrl");
  this->root->addChild(this->model);


  // model tanku pro pocitani kolizi
  this->rootKolize = new SoSeparator;
  rootKolize->addChild(this->rootKolize);

  // posunuti tanku
  this->rootKolize->addChild(this->trans);

  // otoceni tanku
  this->rootKolize->addChild(this->rotace);

  // nacteni modelu tanku
  this->modelKolize = new SoFile;
  this->modelKolize->name.setValue("models/tankkolize.wrl");
  this->rootKolize->addChild(this->modelKolize);
}


class Vybuch
{
  bool aktivni;
  double casStartu;
  int i;
  SbVec3f trojuhelnik[4];
  SoSeparator *root;
  SoSeparator *rootBilboard;
  SoSeparator *otec;
  SoCoordinate3 *coords;
  SoTriangleStripSet *strip;
  SoTranslation *trans;
  SoRotation *rotace;
  SoTexture2 *texture;
  SoTransparencyType *transparencyType;
  SoTextureCoordinate2 *texCoord;

public:
  bool jeAktivni() {return aktivni;}
  void setPosition(const SbVec3f pozice) {trans->translation.setValue(pozice);}
  void setPosition(const float x, const float y, const float z) {trans->translation.setValue(x, y, z);}
  void getPosition(float &x, float &y, float &z) const {trans->translation.getValue().getValue(x, y, z); }
  void start(){aktivni=true;};//odstartuje explozi
  void init(SoSeparator *root);
  void refresh(double cas, const SbRotation & otoceni);
}vybuch;


void Vybuch::init(SoSeparator *root)
{
  i=-1;
  this->aktivni=false;
  this->root=new SoSeparator;
  //this->root->ref();
  this->otec=root;
  this->otec->addChild(this->root);

  this->trans=new SoTranslation;
  this->root->addChild(this->trans);

  this->rotace=new SoRotation;
  this->root->addChild(this->rotace);


  this->rootBilboard=new SoSeparator;
  this->rootBilboard->ref();

  this->transparencyType=new SoTransparencyType;
  transparencyType->value=SoTransparencyType::DELAYED_ADD;
  this->rootBilboard->addChild(this->transparencyType);

  texCoord = new SoTextureCoordinate2;
  this->rootBilboard->addChild(texCoord);
    texCoord->point.set1Value(0, SbVec2f(0, 0));
    texCoord->point.set1Value(1, SbVec2f(0+0.25, 0));
    texCoord->point.set1Value(2, SbVec2f(0, 0+0.25));
    texCoord->point.set1Value(3, SbVec2f(0+0.25, 0+0.25));


  texture = new SoTexture2;
  texture->filename.setValue("maps/explode_0.png");
  texture->model=SoTexture2::REPLACE;
  //texture->blendColor.setValue(0,0,0);
  this->rootBilboard->addChild(texture);

  coords=new SoCoordinate3;
  strip=new SoTriangleStripSet;
  trojuhelnik[0].setValue(5, -5, 0);
  trojuhelnik[1].setValue(-5, -5, 0);
  trojuhelnik[2].setValue(5, 5, 0);
  trojuhelnik[3].setValue(-5, 5, 0);

  coords->point.setValues(0, 4, &trojuhelnik[0]);
  this->rootBilboard->addChild(coords);
  strip->numVertices.set1Value(0, 4);
  this->rootBilboard->addChild(strip);
}


void Vybuch::refresh(double cas, const SbRotation & otoceni)
{
  if(this->aktivni)
  {
    rotace->rotation=otoceni;
    if(i==-1)
    {
      this->root->addChild(this->rootBilboard);
      this->casStartu=cas;
      i=0;
    }

    if(i>15)
    {
      this->root->removeChild(this->rootBilboard);
      aktivni=false;
      i=-1;
      return;
    }

    if(i<=15 && cas>casStartu+i*0.04)
    {
      int dyt=(i%4);
      int dxt=(i/4);
      float dx=dyt*0.25f;
      float dy=0.75f-(dxt*0.25f);
      //printf("%f %f\n", dx, dy);
      texCoord->point.set1Value(0, SbVec2f(dx, dy));
      texCoord->point.set1Value(1, SbVec2f(dx+0.25f, dy));
      texCoord->point.set1Value(2, SbVec2f(dx, dy+0.25f));
      texCoord->point.set1Value(3, SbVec2f(dx+0.25f, dy+0.25f));
      ++i;
    }
  }
}


class Strela
{
  bool aktivni;
  SbVec3f pocatecniPozice;
  double pocatecniCas;
  float smer;
  float sklon;
  float rychlost;
  SoTranslation *trans;
  SoRotation *rotace;
  SoSeparator *root;
  SoSeparator *rodic;
  SoFile *model;
  Tank * tank;
public:
  bool jeAktivni(){return aktivni;}
  void init(SoSeparator *root, Tank * tank);
  void start(double cas, float rychlost, float sklon);
  void stop(double cas);
  void refresh(double cas);
  SbVec3f getPosition() const { return trans->translation.getValue();}
  void getPosition(float &x, float &y, float &z) const {trans->translation.getValue().getValue(x, y, z); }
  void getParam(float &rychlost, float &sklon) const {rychlost=this->rychlost;sklon=this->sklon;}
  ~Strela();
}strela;


void Strela::init(SoSeparator *root, Tank * tank)
{
  this->aktivni=false;

  this->rodic=new SoSeparator;
  root->addChild(this->rodic);

  this->root = new SoSeparator;
  this->root->ref();
  this->tank=tank;

  this->trans = new SoTranslation;
  this->root->addChild(this->trans);
  this->rotace = new SoRotation;
  this->root->addChild(this->rotace);

  // nacteni modelu strely
  this->model = new SoFile;
  this->model->name.setValue("models/strela.wrl");
  this->root->addChild(this->model);
}


//aktivace strely
void Strela::start(double cas, float rychlost, float sklon)
{
  this->aktivni=true;
  this->rodic->addChild(this->root);
  this->rychlost=rychlost;
  this->pocatecniCas=cas;
  this->sklon=sklon;

  // posunuti
  float x, y, z;
  this->tank->getPosition(x, y, z);
  this->smer=tank->getRotation();
  x=x-sinf(this->smer)*4.5f;
  y=2.275f;
  z=z-cosf(this->smer)*4.5f;
  this->pocatecniPozice.setValue(x, y, z);
  this->trans->translation=this->pocatecniPozice;

  // otoceni
  this->rotace->rotation.setValue(SbVec3f(0, 1, 0), this->smer);
}

void Strela::stop(double cas)
{
  if(this->aktivni)
  {
    this->aktivni=false;
    this->rodic->removeChild(this->root);
  }
}


//posunuti v zavislosti na case
void Strela::refresh(double cas)
{
  if(this->aktivni)
  {
    float x, y, z;
    this->pocatecniPozice.getValue(x, y, z);
    cas-=this->pocatecniCas;
    y=y+(float)cas*(rychlost*sinf(this->sklon)-0.5f*(float)cas*gravitace);
    float s=this->rychlost*(float)cas*cosf(this->sklon);
    x=x-sinf(this->smer)*s;
    z=z-cosf(this->smer)*s;
    this->trans->translation.setValue(x, y, z);
  }
}

Strela::~Strela()
{
  if(this->aktivni)
    this->rodic->removeChild(this->root);
  this->root->unref();
}


class StrelaKolize
{
  float x, y, z;
  SbVec3f trojuhelnik[3];// tento trojuhelnik se bude testovat na kolize
  SoCoordinate3 *coords;
  SoTriangleStripSet *strip;
  SoSeparator *rootKolize;
public:
  SoSeparator *root;
  void init(SoSeparator * rootKolize);
  bool test(SbVec3f bod1, SbVec3f bod2);//vrati true pokud dojde ke kolizi, jinak false
}strelaKolize;


void StrelaKolize::init(SoSeparator * rootKolize)
{
  this->rootKolize=rootKolize;
  this->root=new SoSeparator;
  rootKolize->addChild(this->root);
}


//vstupem je minula a aktualni pozice strely
bool StrelaKolize::test(SbVec3f bod1, SbVec3f bod2)
{
  coords=new SoCoordinate3;
  strip=new SoTriangleStripSet;
  trojuhelnik[0]=bod1;
  trojuhelnik[1]=bod2;
  bod2.getValue(x, y, z);
  y-=0.1f;
  trojuhelnik[2].setValue(x, y, z);
  coords->point.setValues(0, 3, &trojuhelnik[0]);
  this->root->addChild(coords);
  strip->numVertices.set1Value(0, 3);
  this->root->addChild(strip);

  kolize=false;
  ida2->apply(this->rootKolize);

  this->root->removeChild(strip);
  this->root->removeChild(coords);

  return kolize;
}


struct Klavesa
{
  enum {
      nahoru = 0, dolu, vlevo, vpravo, strela, zvedniHlaven, sklonHlaven
  };
  SoKeyboardEvent::Key klavesa;
  bool stisknuta;
} klavesy[]= {  {SoKeyboardEvent::UP_ARROW, false},    // pole se stavem klaves
        {SoKeyboardEvent::DOWN_ARROW, false},  //  pouziti: klavesy[Klavesa::dolu].stisknuta
        {SoKeyboardEvent::LEFT_ARROW, false},
        {SoKeyboardEvent::RIGHT_ARROW, false},
        {SoKeyboardEvent::SPACE, false},
        {SoKeyboardEvent::M, false},
        {SoKeyboardEvent::N, false}
        };

void keyboardEvent_cb(void * userdata, SoEventCallback * node)
{
  const SoEvent * event = node->getEvent();

  // obsluha pole "klavesy"
  for(int i=0; i<sizeof(klavesy); ++i)
  {
    if(SoKeyboardEvent::isKeyPressEvent(event, klavesy[i].klavesa))
      klavesy[i].stisknuta=true;

    if(SoKeyboardEvent::isKeyReleaseEvent(event, klavesy[i].klavesa))
      klavesy[i].stisknuta=false;
  }

  // obsluha ESC
  if(SoKeyboardEvent::isKeyPressEvent(event, SoKeyboardEvent::ESCAPE))
#if defined(_WIN32) || defined(__WIN32__)
    SoWin::exitMainLoop();
#else
    SoQt::exitMainLoop();
#endif

  node->setHandled();
}



void prepocitejScenu(double casNovy, double uplynulyCas)
{

#if 1  // vypisovani fps
  static double casfps=casNovy;
  static int fps=0;

  if(casfps+1.0<casNovy)
  {
    ++casfps;
    printf("%i fps\n",fps);
    fps=0;
  }else
    ++fps;
#endif


  float x, y, z;
  float rychlost=tank.getRychlost(); // nova rychlost tanku
  float natoceni=tank.getRotation();
  float natoceniSave=natoceni;  // ulozeni puvodniho natoceni tanku
  SbVec3f poziceSave=tank.getPosition();  // ulozeni puvodni pozice tanku

  // zrychlovani
  if(klavesy[Klavesa::nahoru].stisknuta)
  {
    if(rychlost<maxv)
    {
      rychlost += (float)uplynulyCas*zrychleni;
      if(rychlost>maxv)
        rychlost = maxv;
    }
  }else
  {
    if(rychlost>0)
    {
      rychlost-= (float)uplynulyCas*zpomaleni;
      if(rychlost<0)
        rychlost = 0;
    }
  }

  // zpomalovani
  if(klavesy[Klavesa::dolu].stisknuta)
  {
    if(rychlost>minv)
    {
      rychlost-= (float)uplynulyCas*zrychleni;
      if(rychlost<minv)
        rychlost = minv;
    }
  }else
  {
    if(rychlost<0)
    {
      rychlost += (float)uplynulyCas*zpomaleni;
      if(rychlost>0)
        rychlost = 0;
    }
  }

  // zataceni vlevo
  if(klavesy[Klavesa::vlevo].stisknuta)
  {
    tank.rychlostOt += (float)uplynulyCas*zrychleniOt;
    if(tank.rychlostOt>maxvOt)
      tank.rychlostOt = maxvOt;
  }else
  {
    if(tank.rychlostOt>0)
    {
      tank.rychlostOt-= (float)uplynulyCas*zpomaleniOt;
      if(tank.rychlostOt<0)
        tank.rychlostOt = 0;
    }
  }

  // zataceni vpravo
  if(klavesy[Klavesa::vpravo].stisknuta)
  {
    tank.rychlostOt -= (float)uplynulyCas*zrychleniOt;
    if(tank.rychlostOt<minvOt)
      tank.rychlostOt = minvOt;
  }else
  {
    if(tank.rychlostOt<0)
    {
      tank.rychlostOt+= (float)uplynulyCas*zpomaleniOt;
      if(tank.rychlostOt>0)
        tank.rychlostOt = 0;
    }
  }

  // inverze otaceni pri couvani
  if(rychlost>=0)
    natoceni += (float)uplynulyCas*tank.rychlostOt;
  else
    natoceni -= (float)uplynulyCas*tank.rychlostOt;

  // hlaven nahoru
  if(klavesy[Klavesa::zvedniHlaven].stisknuta)
  {
    tank.rychlostSklonu += (float)uplynulyCas*zrychleniSklonu;
    if(tank.rychlostSklonu>maxvSklonu)
      tank.rychlostSklonu = maxvSklonu;
  }else
  {
    if(tank.rychlostSklonu>0)
    {
      tank.rychlostSklonu-= (float)uplynulyCas*zpomaleniSklonu;
      if(tank.rychlostSklonu<0)
        tank.rychlostSklonu = 0;
    }
  }

  // hlaven dolu
  if(klavesy[Klavesa::sklonHlaven].stisknuta)
  {
    tank.rychlostSklonu -= (float)uplynulyCas*zrychleniSklonu;
    if(tank.rychlostSklonu<minvSklonu)
      tank.rychlostSklonu = minvSklonu;
  }else
  {
    if(tank.rychlostSklonu<0)
    {
      tank.rychlostSklonu+= (float)uplynulyCas*zpomaleniSklonu;
      if(tank.rychlostSklonu>0)
        tank.rychlostSklonu = 0;
    }
  }

  // sklon hlavne
  float sklon=tank.getSklonHlavne();
  sklon += (float)uplynulyCas*tank.rychlostSklonu;

  // omezeni sklonu hlavne podle meznich hodnot
  if(sklon>maxSklonHlavne)
    sklon=maxSklonHlavne;
  else
    if(sklon<minSklonHlavne)
      sklon=minSklonHlavne;

  tank.sklonHlavne=sklon;


  // testuj kolizi mezi "zivymi objekty" (tanky a strely),
  // pokud tank vjede na misto, kde narazil do druheho ziveho objektu,
  // musi se vratit na minulou pozici
  kolize=false;
  ida3->apply(rootkolize);
  if(kolize)
  {
    tank.setPosition(tank.minulaPozice);
    tank.setRotation(tank.minuleNatoceni);
    tank.setRychlost(0.0f);
    rychlost=0.0f;
    natoceniSave=(natoceni=tank.minuleNatoceni);
  }

  // kontrola kolize tanku s bludistem pri rotaci
  if(natoceni!=natoceniSave)
  {
    kolize=false;
    tank.setRotation(natoceni);
    ida->apply(rootkolize);
    if(kolize)
    {
      natoceni=natoceniSave;
      tank.setRotation(natoceni);
    }
  }


  // nova pozice tanku
  float s=(rychlost+tank.getRychlost())*(float)uplynulyCas;  // draha ujeta tankem
  tank.getPosition(x, y, z);

  x-=s*sinf(natoceni);
  z-=s*cosf(natoceni);

  tank.setPosition(x, y, z);
  tank.smer=tank.getPosition()-poziceSave;


  // kontrola kolize tanku s bludistem po posunuti tanku
  kolize=false;
  tmpangle=(float)(M_PI/2.0f);
  tank.KorekceUhlu=0.0f;
  ida->apply(rootkolize);
  if(kolize)
  {
    tank.setPosition(poziceSave);
    tank.getPosition(x,y,z);

    float d=(float)uplynulyCas*sinf(tank.KorekceUhlu)*zpomaleniZed;//zpomaleni o zed
    if(d<0)
      d=-d;
    if(rychlost>0)
    {
      rychlost-=d;
      if(rychlost<0)
        rychlost=0;
    }else
    {
      rychlost+=d;
      if(rychlost>0)
        rychlost=0;
    }

    s=cosf(tank.KorekceUhlu)*(rychlost+tank.getRychlost())*(float)uplynulyCas;  // draha ujeta tankem

    x-=s*sinf(natoceni+tank.KorekceUhlu);
    z-=s*cosf(natoceni+tank.KorekceUhlu);

    tank.setPosition(x, y, z);

    // Pokud i po zpomaleni tanku o zed kolize zustala
    // (napriklad tank narazil v blizkosti rohu a nastala kolize s jinou zdi)
    // => vyresime vracenim tanku na jeho puvodni pozici.
    kolize=false;
    ida->apply(rootkolize);
    if(kolize)
    {
      tank.setPosition(poziceSave);
    }
  }

  // ulozeni nove rychlosti
  tank.setRychlost(rychlost);

  // ulozeni hodnot pro pripadny navrat(pokud pojede na stejne misto jako jiny tank)
  tank.minulaPozice=tank.getPosition();
  tank.minuleNatoceni=tank.getRotation();


  // nastaveni kamery podle polohy, pozice a rychlosti tanku
  SbRotation otockameru, sklopkameru;
  otockameru.setValue(SbVec3f(0, 1, 0), natoceni);
  float sklopeni=-.38f; // sklopeni kamery
  float posun=18;    // posun kamery
  float vyska=11.0f; // vyska umisteni kamery
  if(rychlost<0)
  {
    posun=18.0f-rychlost*0.44f;  // posun kamery v zavislosti na rychlosti couvani
  }
  sklopkameru.setValue(SbVec3f(1, 0, 0), sklopeni);
  kamera->orientation=sklopkameru*otockameru;

  tank.getPosition(x, y, z);
  x+=posun*sinf(natoceni);
  y=vyska;
  z+=posun*cosf(natoceni);
  kamera->position.setValue(x, y, z);

  static SbVec3f pozice1, pozice2;//stara a nova pozice strely

  if(strela.jeAktivni())
  {
    strela.refresh(casNovy);
    pozice2=strela.getPosition();
    strelaKolize.test(pozice1, pozice2);

    pozice2.getValue(x, y, z);
    if(kolize||y<0)
    {
      vybuch.setPosition(pozice1);
      vybuch.start();
      strela.stop(casNovy);
    }
    pozice1=pozice2;
  }

  if(klavesy[Klavesa::strela].stisknuta)
  {
    if(!strela.jeAktivni())
    {
      static double casSave=-10;
      if(casSave+1.0<casNovy)//sekunda mezi vystrely
      {
        klavesy[Klavesa::strela].stisknuta=false;
        strela.start(casNovy, 400.f, tank.getSklonHlavne());
        casSave=casNovy;
      }
      pozice1=strela.getPosition();
    }
  }
  vybuch.refresh(casNovy, kamera->orientation.getValue());
}


void simulationStep(void *data, SoSensor *sensor)
{
  // spotej as od minulho voln simulationStep
  static double casNovy=SbTime::getTimeOfDay().getValue(), casStary;

  casStary=casNovy;
  casNovy=SbTime::getTimeOfDay().getValue();
  double uplynulyCas=casNovy-casStary;

  // aktualizuj scnu
  prepocitejScenu(casNovy, uplynulyCas);

  // rescheduluj sensor (zpsob nov zavoln tto funkce)
  sensor->schedule();
}


static SoIntersectionDetectionAction::Resp
intersectionCB(void * closure, const SoIntersectingPrimitive * pr1,  const SoIntersectingPrimitive * pr2)
{
  SbPlane *plocha;
  if( pr2->path->containsNode(tank.rootKolize) )
    plocha=new SbPlane(pr1->vertex[0], pr1->vertex[1], pr1->vertex[2]);
  else
    plocha=new SbPlane(pr2->vertex[0], pr2->vertex[1], pr2->vertex[2]);

  SbVec3f normala;
  float x,y,z;
  normala=plocha->getNormal();
  //normala.getValue(x,y,z);

  if(tank.smer.length()>0)
  {
    SbRotation rot(normala, tank.smer);
    SbVec3f axis;
    float angle;
    rot.getValue(axis, angle);

    //printf("angle %f \n", angle);
    if(angle>tmpangle && angle>M_PI/2.0)
    {
      tmpangle=angle;
      axis.getValue(x,y,z);
      tank.KorekceUhlu=(float)(-y*(angle-M_PI/2.0f));
      //printf("angle %f %f\n", angle, tank.KorekceUhlu);
    }
  }

  kolize=true;
  return SoIntersectionDetectionAction::NEXT_PRIMITIVE;
}


// Funkce filtruje objekty, a rika, mezi kterymi objekty se bude kolize detekovat.
// Tato filtrovaci funkce zabranuje detekovat kolize mezi jednotlivymi castmi bludiste.
SbBool filterCB(void * closure, const SoPath * p1, const SoPath * p2)
{
  if( (p1->containsNode (bludisteKolize) && p2->containsNode (bludisteKolize)) )
    return FALSE;
  return TRUE;
}


bool testTank2;
static SoIntersectionDetectionAction::Resp
intersectionCB2(void * closure, const SoIntersectingPrimitive * pr1, const SoIntersectingPrimitive * pr2)
{
  //printf("sopath %f %f %f\n",x,y,z);
  kolize=true;

  //zasah tanku
  if(testTank2)
  {
    ++tank.skore;
    return SoIntersectionDetectionAction::ABORT;
  }

    return SoIntersectionDetectionAction::NEXT_PRIMITIVE;
}


// Funkce filtruje objekty, a rika, mezi kterymi objekty se bude kolize detekovat.
// Tato filtrovaci funkce zabranuje detekovat kolize mezi jednotlivymi castmi bludiste.
SbBool filterCB2(void * closure, const SoPath * p1, const SoPath * p2)
{
  bool containsStrela=p1->containsNode(strelaKolize.root)||p2->containsNode(strelaKolize.root);
  bool containsTank=p1->containsNode(tank.rootKolize)||p2->containsNode(tank.rootKolize);
  if( p1->containsNode (bludisteKolize) && p2->containsNode (bludisteKolize) ||
      containsStrela && containsTank)
    return FALSE;
  return TRUE;
}


//kolize mezi tanky
static SoIntersectionDetectionAction::Resp
intersectionCB3(void * closure, const SoIntersectingPrimitive * pr1, const SoIntersectingPrimitive * pr2)
{
  kolize=true;
  return SoIntersectionDetectionAction::ABORT;
}


// Funkce filtruje objekty, a rika, mezi kterymi objekty se bude kolize detekovat.
// Tato filtrovaci funkce povoluje detekovat kolize mezi vsim,
// co neni bludiste.
SbBool filterCB3(void * closure, const SoPath * p1, const SoPath * p2)
{
  if(p1->containsNode (bludisteKolize) || p2->containsNode (bludisteKolize))
    return FALSE;
  return TRUE;
}


int main(int , char **argv)
{
  printf("\nVitejte ve hre Tank Hunters\n\n");

  printf("*************************************************************\n");
  printf("* pohyb tanku: sipkami                                      *\n");
  printf("* strileni: mezernikem                                      *\n");
  printf("* sklapeni hlavne: M a N                                    *\n");
  printf("*************************************************************\n");


  // Inicializuj Inventor
#if defined(_WIN32) || defined(__WIN32__)
  HWND window = SoWin::init(argv[0]);
#else
  QWidget *window = SoQt::init(argv[0]);
#endif
  if (window == NULL) exit(1);

  // vytvo koen scny
  root = new SoSeparator;
  root->ref();


  // Vlo kameru do scny
  kamera = new SoPerspectiveCamera;
  kamera->nearDistance = 4;
  kamera->farDistance = 4096;
  root->addChild(kamera);

  // nastaveni modelu osvetleni
  SoLightModel *lmodel = new SoLightModel;
  lmodel->model.setValue(SoLightModel::BASE_COLOR);
  root->addChild(lmodel);

  // textury z WRML soubor bodou natny okamit, nikoliv na pozad
  SoVRMLImageTexture::setDelayFetchURL(FALSE);

  // bludiste
  SoFile *model = new SoFile;
  model->name.setValue("models/bludiste.wrl");
  root->addChild(model);


  // vytvo koen scny pro potn koliz
  rootkolize = new SoSeparator;
  rootkolize->ref();

  // bludiste pro pocitani kolizi
  bludisteKolize = new SoSeparator;
  rootkolize->addChild(bludisteKolize);

  // model bludiste pro pocitani kolizi
  SoFile *modelKolize = new SoFile;
  modelKolize->name.setValue("models/bludistekolize.wrl");
  bludisteKolize->addChild(modelKolize);


  // inicializace tanku
  tank.init(root, rootkolize);
  tank.setPosition(10,0,-30);
  tank.setRotation(3.1f);

  strela.init(root, &tank);
  vybuch.init(root);


  // kolize
  ida = new SoIntersectionDetectionAction;
  ida->setFilterCallback(filterCB, NULL);
  ida->addIntersectionCallback(intersectionCB, NULL);
  ida->setManipsEnabled(FALSE);
  ida->setDraggersEnabled(FALSE);

  ida2 = new SoIntersectionDetectionAction;
  ida2->setFilterCallback(filterCB2, NULL);
  ida2->addIntersectionCallback(intersectionCB2, NULL);
  ida2->setManipsEnabled(FALSE);
  ida2->setDraggersEnabled(FALSE);

  ida3 = new SoIntersectionDetectionAction;
  ida3->setFilterCallback(filterCB3, NULL);
  ida3->addIntersectionCallback(intersectionCB3, NULL);
  ida3->setManipsEnabled(FALSE);
  ida3->setDraggersEnabled(FALSE);

  strelaKolize.init(rootkolize);


  // Vytvor renderovaci okno
#if defined(_WIN32) || defined(__WIN32__)
  SoWinRenderArea *renderArea = new SoWinRenderArea(window);
#else
  SoQtRenderArea *renderArea = new SoQtRenderArea(window);
#endif
  //renderArea->setTransparencyType(SoGLRenderAction::DELAYED_ADD);

  // Nastav renderovac okno
  renderArea->setSceneGraph(root);
  renderArea->setTitle("Tank Hunters");
  renderArea->show();

  // Obsluha klavesnice
  SoEventCallback * cb = new SoEventCallback;
  cb->addEventCallback(SoKeyboardEvent::getClassTypeId(), keyboardEvent_cb, NULL);
  root->insertChild(cb, 0);

  // Sensor pro nekonen voln simulationStep
  SoOneShotSensor * sensor =new SoOneShotSensor(simulationStep, NULL);
  sensor->schedule();

  // Zobraz okna a rozje renderovac smyku
#if defined(_WIN32) || defined(__WIN32__)
  SoWin::show(window);
  SoWin::mainLoop();
#else
  SoQt::show(window);
  SoQt::mainLoop();
#endif


  // Uvolni pamet
  delete renderArea;
  root->unref();

  return 0;
}
